# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import pytest

from ..pool import pool
from .test_cli import CLISubCommandTests


@pytest.fixture
def subcmd():
    return "backup"


class TestBackup(CLISubCommandTests):
    def test_default_all_managers_output_to_console(self, invoke, subcmd):
        result = invoke(subcmd)
        assert result.exit_code == 0
        assert "Backup package list to <stdout>" in result.stderr
        self.check_manager_selection(result)

    def test_output_to_console(self, invoke, subcmd):
        result = invoke(subcmd, "-")
        assert result.exit_code == 0
        assert "Backup package list to <stdout>" in result.stderr
        self.check_manager_selection(result)

    def test_output_to_file(self, invoke, subcmd):
        result = invoke(subcmd, "mpm-packages.toml")
        assert result.exit_code == 0
        assert "mpm-packages.toml" in result.stderr
        self.check_manager_selection(result)

    @pytest.mark.parametrize("mid", pool.default_manager_ids)
    def test_single_manager_file_output(self, mid, invoke, subcmd):
        result = invoke("--manager", mid, subcmd, "mpm-packages.toml")
        assert result.exit_code == 0
        assert "mpm-packages.toml" in result.stderr
        self.check_manager_selection(result, {mid})



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import json
from string import ascii_lowercase, digits

import pytest
from boltons.iterutils import same

from ..pool import pool
from .test_cli import CLISubCommandTests, CLITableTests


@pytest.fixture
def subcmd():
    return "outdated"


XBAR_KEYWORDS = frozenset({"shell"}.union({f"param{i}" for i in range(1, 10)}))


class TestOutdated(CLISubCommandTests, CLITableTests):
    @pytest.mark.parametrize("mid", pool.default_manager_ids)
    def test_single_manager(self, invoke, mid, subcmd):
        result = invoke("--manager", mid, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, {mid})

    def test_json_parsing(self, invoke, subcmd):
        result = invoke("--output-format", "json", subcmd)
        assert result.exit_code == 0
        data = json.loads(result.stdout)

        assert data
        assert isinstance(data, dict)
        assert set(data).issubset(pool.default_manager_ids)

        for manager_id, info in data.items():
            assert isinstance(manager_id, str)
            assert isinstance(info, dict)

            assert isinstance(info["id"], str)
            assert isinstance(info["name"], str)

            keys = {"errors", "id", "name", "packages"}
            if "upgrade_all_cli" in info:
                assert isinstance(info["upgrade_all_cli"], str)
                keys.add("upgrade_all_cli")
            assert set(info) == keys

            assert isinstance(info["errors"], list)
            if info["errors"]:
                assert same(map(type, info["errors"]), str)

            assert info["id"] == manager_id

            assert isinstance(info["packages"], list)
            for pkg in info["packages"]:
                assert isinstance(pkg, dict)

                assert set(pkg) == {
                    "id",
                    "installed_version",
                    "latest_version",
                    "name",
                    "upgrade_cli",
                }

                assert isinstance(pkg["id"], str)
                assert isinstance(pkg["installed_version"], str)
                assert isinstance(pkg["latest_version"], str)
                assert isinstance(pkg["name"], str)
                assert isinstance(pkg["upgrade_cli"], str)

    def test_cli_format_plain(self, invoke, subcmd):
        result = invoke("--output-format", "json", subcmd, "--cli-format", "plain")
        for outdated in json.loads(result.stdout).values():
            for infos in outdated["packages"]:
                assert isinstance(infos["upgrade_cli"], str)

    def test_cli_format_fragments(self, invoke, subcmd):
        result = invoke("--output-format", "json", subcmd, "--cli-format", "fragments")
        for outdated in json.loads(result.stdout).values():
            for infos in outdated["packages"]:
                assert isinstance(infos["upgrade_cli"], list)
                assert same(map(type, infos["upgrade_cli"]), str)

    def test_cli_format_xbar(self, invoke, subcmd):
        result = invoke("--output-format", "json", subcmd, "--cli-format", "xbar")
        for outdated in json.loads(result.stdout).values():
            for infos in outdated["packages"]:
                assert isinstance(infos["upgrade_cli"], str)
                assert "param1=" in infos["upgrade_cli"]
                for param in infos["upgrade_cli"].split(" | "):
                    k, v = param.split("=", 1)
                    assert k in XBAR_KEYWORDS
                    assert set(v.lower()).issubset(
                        digits + ascii_lowercase + './-_+="\\@:'
                    )



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

from click_extra.platform import LINUX

from ..base import PackageManager
from ..version import parse_version


class DNF(PackageManager):

    """
    Documentation: https://dnf.readthedocs.io/en/latest/command_ref.html
    """

    platforms = frozenset({LINUX})

    requirement = "4.0.0"

    cli_names = ("dnf", "yum")

    """
    .. code-block:: shell-session

        ► dnf --version
        4.9.0
    """

    pre_args = ("--color=never",)

    list_cmd_regexp = re.compile(r"(\S+)\.\S+\s+(\S+)\s+\S+")

    @property
    def installed(self):
        """
        .. code-block:: shell-session

            ► dnf --color=never list --installed
            Installed Packages
            acl.x86_64               2.2.53-1.el8                          @anaconda
            audit.x86_64             3.0-0.10.20180831git0047a6c.el8       @anaconda
            audit-libs.x86_64        3.0-0.10.20180831git0047a6c.el8       @anaconda
            (...)
        """
        installed = {}

        output = self.run_cli("list", "--installed")

        for package in output.splitlines()[1:]:
            match = self.list_cmd_regexp.match(package)
            if match:
                package_id, installed_version = match.groups()
                installed[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "installed_version": parse_version(installed_version),
                }

        return installed

    @property
    def outdated(self):
        """
        .. code-block:: shell-session

            ► dnf --color=never list --upgrades
            Last metadata expiration check: 0:22:12 ago on Sun 03 Apr 2022.
            Available Upgrades
            acl.x86_64               2.2.53-1.el8                          updates
            audit.x86_64             3.0-0.10.20180831git0047a6c.el8       updates
            audit-libs.x86_64        3.0-0.10.20180831git0047a6c.el8       updates
            (...)
        """
        outdated = {}

        output = self.run_cli("list", "--upgrades")

        for package in output.splitlines()[2:]:
            match = self.list_cmd_regexp.match(package)
            if match:
                package_id, latest_version = match.groups()
                outdated[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": parse_version(latest_version),
                }

        return outdated

    def search(self, query, extended, exact):
        """
        .. code-block:: shell-session

            ► dnf --color=never search usd
            Last metadata expiration check: 0:06:37 ago on Sun 03 Apr 2022.
            =================== Name Exactly Matched: usd =====================
            usd.aarch64 : 3D VFX pipeline interchange file format
            =================== Name & Summary Matched: usd ===================
            python3-usd.aarch64 : Development files for USD
            usd-devel.aarch64 : Development files for USD
            ======================= Name Matched: usd =========================
            lvm2-dbusd.noarch : LVM2 D-Bus daemon
            usd-libs.aarch64 : Universal Scene Description library
        """
        matches = {}

        output = self.run_cli("search", query)

        regexp = re.compile(r"(\S+)\.\S+\s:\s(\S+)")

        for line in output.splitlines()[1:]:
            # Skip section headers.
            if line.startswith("="):
                continue

            # Extract package ID and description.
            match = regexp.match(line)
            if match:
                package_id, description = match.groups()

                # Any line matches in extended mode.
                if not extended:

                    # Skip package if ID does not exactly match the query in exact mode.
                    if exact:
                        if package_id != query:
                            continue
                    # Skip package if query is not found in ID.
                    else:
                        if query not in package_id:
                            continue

                matches[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": None,
                }

        return matches

    def install(self, package_id):
        """
        .. code-block:: shell-session

            ► sudo dnf --color=never --assumeyes install pip
        """
        super().install(package_id)
        return self.run_cli(
            "--assumeyes", "install", package_id, override_pre_cmds=("sudo",)
        )

    def upgrade_cli(self, package_id=None):
        """
        .. code-block:: shell-session

            ► sudo dnf --color=never --assumeyes upgrade pip

        .. code-block:: shell-session

            ► sudo dnf --color=never --assumeyes upgrade
        """
        return self.build_cli("upgrade", package_id)

    def sync(self):
        """
        .. code-block:: shell-session

            ► dnf --color=never check-update
        """
        super().sync()
        self.run_cli("check-update")

    def cleanup(self):
        """
        .. code-block:: shell-session

            ► sudo dnf --color=never --assumeyes autoremove
            ► dnf --color=never clean all
        """
        super().cleanup()
        self.run_cli("--assumeyes", "autoremove", override_pre_cmds=("sudo",))
        self.run_cli("clean", "all")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

from click_extra.platform import LINUX

from ..base import PackageManager
from ..version import TokenizedString, parse_version


class OPKG(PackageManager):

    platforms = frozenset({LINUX})

    requirement = "0.2.0"

    version_regex = r"opkg\s+version\s+(?P<version>\S+)"
    """
    .. code-block:: shell-session

        ► opkg --version
        opkg version 0.3.6 (libsolv 0.7.5)
    """

    @property
    def installed(self):
        """Fetch installed packages from ``opkg list-installed`` output.

        Raw CLI output samples:

        .. code-block:: shell-session
            ► opkg list-installed
            3rd-party-feed-configs - 1.1-r0
            aio-grab - 1.0+git71+c79e264-r0
            alsa-conf - 1.1.9-r0
            alsa-state - 0.2.0-r5
            alsa-states - 0.2.0-r5
            alsa-utils-alsactl - 1.1.9-r0
            avahi-daemon - 0.7-r0
            base-files - 3.0.14-r89
            base-files-dev - 3.0.14-r89
            base-passwd - 3.5.29-r0
            bash - 5.0-r0
            bash-completion - 2.9-r0
            bash-completion-dev - 2.9-r0
            bash-dev - 5.0-r0
            binutils - 2.32.0-r0
            busybox - 1.31.0-r0
            busybox-inetd - 1.31.0-r0
            busybox-mdev - 1.31.0-r0
            busybox-syslog - 1.31.0-r0
            busybox-udhcpc - 1.31.0-r0
        """
        installed = {}

        output = self.run_cli("list-installed")

        regexp = re.compile(r"(\S+) - (\S+)")
        for package in output.splitlines():
            match = regexp.match(package)
            if match:
                package_id, installed_version = match.groups()
                installed[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "installed_version": parse_version(installed_version),
                }

        return installed

    @property
    def outdated(self):
        """Fetch outdated packages from ``opkg list-upgradable`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► opkg list-upgradable
            openpli-bootlogo - 20190717-r0 - 20190718-r0
            enigma2-hotplug - 2.7+git1720+55c6b34-r0 - 2.7+git1722+daf2f52-r0
        """
        outdated = {}

        output = self.run_cli("list-upgradable")

        regexp = re.compile(r"(\S+) - (\S+) - (\S+)")
        for package in output.splitlines():
            match = regexp.match(package)
            if match:
                package_id, installed_version, latest_version = match.groups()
                outdated[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": parse_version(latest_version),
                    "installed_version": parse_version(installed_version),
                }

        return outdated

    def search(self, query, extended, exact):
        """Simulate search by listing all packages.

        .. code-block:: shell-session

            ► opkg list
        """
        matches = {}

        # opkg doesn't have a working 'search', so get all packages and
        # filter the packages later.
        output = self.run_cli("list")

        regexp = re.compile(
            r"""
            (?P<package_id>\S+)
            \ -\
            (?P<version>\S+)
            \ -\
            (?P<description>.+)
            """,
            re.VERBOSE | re.MULTILINE,
        )

        for package_id, version, description in regexp.findall(output):
            # Skip all non-stricly matching package IDs in exact mode.
            if exact:
                if query != package_id:
                    continue
            else:
                # All other modes search for matching in package IDs and
                # names.
                searched_content = set(map(str, TokenizedString(package_id)))

                # Also search within the description in extended mode.
                if extended:
                    searched_content.update(set(map(str, TokenizedString(description))))

                # Skip package if not all sub-strings are present in the
                # searched content.
                query_parts = set(map(str, TokenizedString(query)))
                if not query_parts.issubset(searched_content):
                    continue

            matches[package_id] = {
                "id": package_id,
                "name": package_id,
                "latest_version": parse_version(version),
            }

        return matches

    def install(self, package_id):
        """Install one package.

        .. code-block:: shell-session

            ► opkg install enigma2-hotplug
        """
        super().install(package_id)
        return self.run_cli("install", package_id)

    def upgrade_cli(self, package_id=None):
        """
        .. code-block:: shell-session

            ► opkg upgrade enigma2-hotplug
        """
        return self.build_cli(
            "upgrade",
            package_id,
        )

    def sync(self):
        """
        .. code-block:: shell-session

            ► opkg update
        """
        super().sync()
        self.run_cli("update")
